/**
 * @ingroup   winamp68
 * @file      winamp68.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      06/11/2000
 * @brief     sc68 input plugin for winamp.
 * @version   $Id: winamp68.c,v 1.1.1.1 2001/12/24 20:50:11 ben Exp $
 */

/*
 *                       sc68 - winamp input plugin
 *         Copyright (C) 2001 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "sc68glabout.h"

#include <windows.h>
#include <mmreg.h>
#include <msacm.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "resource.h"

#include "in2.h"
#include "version68.h"


/* 512 Kb 68K memory buffer */
#define MEMORY (1<<19)

//#define WANT_BROWSEINFO_68
//#define _WRAPPER_

/*
* SC68 includes
*/
#include "sc68app/SC68.h"
#include "sc68app/SC68app.h"
#include "file68/SC68error.h"
#include "file68/SC68utils.h"
#include "file68/SC68file.h"
#ifdef WANT_BROWSEINFO_68
# include "browse68/SC68browse.h"
#endif
#include "file68/SC68os_general.h"
#include "sc68app/SC68os_interface.h"
#include "sc68app/SC68play.h"
#include "sc68app/SC68delay.h"
#include "sc68app/SC68mixer.h"

/*
* Emulator includes
*/
#include "emu68/emu68.h"
#include "emu68/ioplug68.h"
#include "io68/io68.h"

static SC68app_t app;
// post this to the main window at end of file (after playback as stopped)
#define WM_WA_MPEG_EOF WM_USER+2

// raw configuration
//#define NCH 2
//#define SAMPLERATE 44100
//#define BPS 16

In_Module mod; // the output module (declared near the bottom of this file)
static HINSTANCE about_dll=0;

// sample buffer [ get more room
static char * sample_buffer = 0;
static int sample_buffer_sz = 0;

HANDLE input_file=INVALID_HANDLE_VALUE; // input file handle

static volatile int killDecodeThread=0;					// the kill switch for the decode thread
HANDLE thread_handle=INVALID_HANDLE_VALUE;	// the handle to the decode thread

DWORD WINAPI __stdcall DecodeThread(void *b); // the decode thread procedure

int CreateFileInfoDialog(const char *fn, HWND hwnd);
int CreateConfigureDialog(HWND hwnd, SC68app_t *);
//int CreateAboutDialog(int rundemo, HWND hwnd, SC68app_t *app);
static int (*CreateAboutDialogProc)(int rundemo, sc68glabout_parm_t *p);

#ifndef _DEBUG
# define ABOUT_DLL_NAME "sc68glabout.dll"
#else
# define ABOUT_DLL_NAME "sc68glaboutd.dll"
#endif

/* Display to output debug
 * statcked error messages
 */
static void spool_error_message(void)
{
#ifdef _DEBUG
  char *s;

  if (s = SC68error_get(), s != NULL) {
    SC68os_pdebug("Stacked Error Message:\n");
    do {
      SC68os_pdebug("%s\n", s);
    } while (s = SC68error_get(), s != NULL);
  } else {
    SC68os_pdebug("No stacked error message\n");
  }
#endif

}

/* Update and save configuration file */
static void save_config(void)
{
  int i;
  if (i = SC68config_get_id("total_time"), i >= 0)
    app.config[i] = app.time.total;
  if (i = SC68config_get_id("version"), i >= 0)
    app.config[i] = app.version;

  SC68config_save(&app.config,
                  SC68os_resource_name(SC68rsc_config, "config"));
}

/* Initialize 68K stuff * : alloc memory
 */
static int init_68K(reg68_t * r, u32 sz)
{
  if (EMU68_init(sz) < 0)
    return SC68error_add(EMU68error_get());
  *r = reg68;
  r->sr = 0x2000;
  r->a[7] = sz - 4;
  if (YM_init() < 0)
    return SC68error_add("Yamaha 2149 emulator init failed");
  if (MW_init() < 0)
    return SC68error_add("Microwire emulator init failed");
  if (MFP_init() < 0)
    return SC68error_add("MFP emulator init failed");
  if (PL_init() < 0)
    return SC68error_add("Paula emulator init failed");
  return 0;
}

/* Set sc68app struct to default value
 */
static void set_sc68app_default(SC68app_t * app)
{
  memset(app, 0, sizeof(*app));
  app->cur_track = -1;
  SC68config_default(&app->config);
}

// avoid CRT. Evil. Big. Bloated.
BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
	return TRUE;
}


void config(HWND hwndParent)
{
  CreateConfigureDialog(hwndParent, &app);
}

void about(HWND hwndParent)
{
  static char txt[] =
#ifdef _DEBUG
  "DEBUG VERSION of Atari-ST & Amiga soundchip emulator v" WINAMP68_VERSION_STR " by BeN(jamin) Gerard / sashipa";
#else
  "Atari-ST & Amiga soundchip emulator v" WINAMP68_VERSION_STR " by BeN(jamin) Gerard / sashipa";
#endif

  if (CreateAboutDialogProc) {
    sc68glabout_parm_t p;
    p.hwnd=hwndParent ;    // Parent window
    p.in_mod=&mod ;  // winamp input modules
    p.paula= paula;   // paula register buffer adress
    p.ymread= YM_readreg;  // pointer to YM_read() function
    p.app=&app ;     // sc68app_t *
    p.rscname= SC68os_resource_name; // pointer to SC68os_resource_name()

    CreateAboutDialogProc(0, &p);
  } else {
    MessageBox(hwndParent, txt, "about sc68", MB_OK);
  }
}

static unsigned int TimeToTrack(disk68_t *d, int n)
{
  int i;
  unsigned int cnt = 0;
  if (n >= d->nb_six) {
    n = d->nb_six - 1;
  }
  for (i=0; i<n; ++i) {
    cnt += d->mus[i].time * 1000;
  }
  return cnt;
}

/* any one-time initialization goes here (configuration reading, etc) */
void init()
{
  int r, frq, hres;
  char execname[512];

  SC68os_pdebug("\nINIT XMMS68 pluggin\n");

  hres = GetModuleFileName((HMODULE) mod.hDllInstance, execname, sizeof(execname));
  if(hres<0) {
    execname[0] = 0;
  }

  /* check whether EMU68 is compiled without debug option */
  if (EMU68_sizeof_reg68() != sizeof(reg68_t)) {
    SC68error_add
        ("Internal : Bad EMU68 version : should not define SC68DEBUG ...");
    return;
  }

  /* Set default configuration */
  set_sc68app_default(&app);

  /* Get exec path : for resource */
  if (SC68os_get_exec_path(execname) < 0) {
    return;
  }

  /* Get the config file OR create if it doesn't exist */
  SC68os_pdebug("Read config file ...\n");
  if (SC68config_load(&app.config,
                      SC68os_resource_name(SC68rsc_config,
                                           SC68_DEFAULT_CONFIG_FILE)) < 0) {
    SC68error_get();
    SC68error_get();
    SC68config_save(&app.config,
                    SC68os_resource_name(SC68rsc_config,
                                         SC68_DEFAULT_CONFIG_FILE));
  }
  SC68os_pdebug("OK\n");

  SC68config_display(&app.config);

  /* Open application ( OS dependant code ) */
  r = SC68os_interface_open(execname, &app);

  /* Open error */
  if (r < 0) {
    return;
  }

  /* Copy config parameters to app */
  SC68app_configure(&app);

  /* Delay init : from config file parameters ( SC68app_configure() ) */
  app.delay.buffer_size = SC68delay_init(app.delay.buffer_size);
  if (app.delay.buffer_size < 0)
    return;

  frq = (int) app.mix.real_frq;

  /* First instance : Do mucho initialization */
  if (r > 0) {
#ifdef WANT_BROWSEINFO_68
    /* Browse Info Data Base init */
    if (SC68browse_init(SC68_DEFAULT_BIDB_FILE) < 0) {
      return;
    }
    SC68os_pdebug("Browse init : OK\n");
#endif

    /* 68K & IO-Emulators init */
    if (init_68K(&app.reg68, MEMORY)) {
      spool_error_message();
      return;
    }
    SC68os_pdebug("68K emulator init : OK\n");

    /* IO replay frequency init */
    app.mix.real_frq = 44100;
    YM_set_replay_frq(app.mix.real_frq);
    MW_set_replay_frq((int) app.mix.real_frq);
    PL_set_replay_frq((int) app.mix.real_frq);
    app.mix.buf = YM_get_buffer();
  }

#ifdef _DEBUG
  /* $$$ Debug mode runs always about window ! */
  app.version = 0;
#endif

  about_dll = LoadLibrary(SC68os_resource_name(SC68rsc_dll, ABOUT_DLL_NAME));
  if (about_dll) {
    CreateAboutDialogProc = (int (*)(int, sc68glabout_parm_t *))GetProcAddress(about_dll, "CreateAboutDialog");
  }

  if (app.version != WINAMP68_VERSION) {

    app.version = WINAMP68_VERSION;
    save_config(); /* Prevent About to crash winamp at each launch if About dialog bugs ! */
    if (CreateAboutDialogProc) {
      sc68glabout_parm_t p;
      p.hwnd=mod.hMainWindow;    // Parent window
      p.in_mod=&mod ;  // winamp input modules
      p.paula= paula;   // paula register buffer adress
      p.ymread= YM_readreg;  // pointer to YM_read() function
      p.app=&app ;     // sc68app_t *
      p.rscname= SC68os_resource_name; // pointer to SC68os_resource_name()
      CreateAboutDialogProc (1, &p);
    }
  }
}

/* one-time deinit, such as memory freeing */
void quit()
{
  SC68os_pdebug("\nCLEAN UP winamp68 pluggin\n");
  SC68os_pdebug("EMU68 : Clean Exit\n");
  EMU68_kill();
  SC68os_interface_close(&app);
#ifdef WANT_BROWSEINFO_68
  if (app.browse_info)
    SC68browse_write(SC68_DEFAULT_BIDB_FILE);
#endif
  save_config();
  SC68os_pdebug("\nSC68 : Clean Exit\n");
  spool_error_message();
  if (about_dll) {
    FreeLibrary(about_dll);
    about_dll = 0;
  }
}

int isourfile(char *fn)
{
  return (SC68app_verify_diskfile(fn) == 0);
}
// used for detecting URL streams.. unused here. strncmp(fn,"http://",7) to detect HTTP streams, etc

int play(char *filename)
{
  disk68_t *d = 0;
	int maxlatency;
	int thread_id;

  /* Stop sc68 processing */
  SC68app_desactiv(&app);
	app.pause=0;

  /* Load disk and generate browse info */
#ifdef WANT_BROWSEINFO_68
  d = SC68app_load_diskfile(SC68os_absolute_fname(filename), app.browse_info);
#else
  d = SC68app_load_diskfile(SC68os_absolute_fname(filename), 0);
#endif
  if (!d) {
    goto error;
  }

  /* Set as current application disk, load default track.
   * If app.force_track == 0, disk default track is loaded, else app.force_track-1 is loaded
   */
  if (SC68load(&app, d) < 0) {
    goto error;
  }

  /* Set default track to first ... */
  /*
  if (SC68track(&app, 0) < 0) {
    goto error;
  }
  */

  /* Start current track */
  if (SC68play(&app) < 0) {
    goto error;
  }

  /* Open output pluggin */
	maxlatency = mod.outMod->Open(44100,2,16, -1,-1);
	if (maxlatency < 0) {
    goto error;
	}

	// dividing by 1000 for the first parameter of setinfo makes it
	// display 'H'... for hundred.. i.e. 14H Kbps.
	mod.SetInfo((44100*16*2)/1000,44100/1000,2,1);

	// initialize vis stuff
	mod.SAVSAInit(maxlatency,44100);
//	mod.VSASetInfo(2,44100);
	mod.VSASetInfo(44100,2);
	//mod.outMod->SetVolume(-666); // set the output plug-ins default volume

  /* Correct winamp time slider, which is synchronized on output module */
  if (mod.outMod) {
    mod.outMod->Flush(TimeToTrack(app.cur_disk, app.cur_track));
  }

	killDecodeThread=0;
	thread_handle = (HANDLE) CreateThread(NULL,0,(LPTHREAD_START_ROUTINE) DecodeThread,(void *) &killDecodeThread,0,&thread_id);

  SC68app_activ(&app);


	return 0;

error:
  if (d) {
    SC68_free(d);
  }
  SC68app_desactiv(&app);;
  spool_error_message();
  return 1;
}

void pause()
{
  if(!app.pause) {
    SC68pause(&app);
  }
  mod.outMod->Pause(1);
}

void unpause()
{
  if(app.pause) {
    SC68pause(&app);
  }
  mod.outMod->Pause(0);
}

int ispaused()
{
  return app.pause;
}

void stop()
{
  killDecodeThread = 1;
	if (thread_handle != INVALID_HANDLE_VALUE)
	{
		if (WaitForSingleObject(thread_handle,INFINITE) == WAIT_TIMEOUT)
		{
			MessageBox(mod.hMainWindow,"error asking thread to die!\n","error killing decode thread",0);
			TerminateThread(thread_handle,0);
		}
		CloseHandle(thread_handle);
		thread_handle = INVALID_HANDLE_VALUE;
	}
  SC68app_eject(&app);
	mod.SAVSADeInit();
	if (mod.outMod) mod.outMod->Close();
}

int getlength()
{
  return app.cur_disk ? app.cur_disk->time * 1000 : 0;
}

int getoutputtime() {
	return mod.outMod->GetOutputTime();
}

void setoutputtime(int time)
{
  if (app.ready && app.cur_disk) {
    disk68_t *d = app.cur_disk;
    int i, cnt;

    for (i = cnt = 0; i < d->nb_six; ++i) {
      int cnt2;

      cnt2 = cnt + d->mus[i].time * 1000;
      if (time >= cnt && time <= cnt2) {
        if (i != app.cur_track) {
          SC68track(&app, i);
          if (mod.outMod) {
            mod.outMod->Flush (cnt);
          }
        }
      }
      cnt = cnt2;
    }
  }
}

void setvolume(int volume)
{
  mod.outMod->SetVolume(volume);
}

void setpan(int pan)
{
  mod.outMod->SetPan(pan);
}

int infoDlg(char *fn, HWND hwnd)
{
  return CreateFileInfoDialog(fn, hwnd);
}

void getfileinfo(char *filename, char *title, int *length)
{
  disk68_t *d;

  if (title) {
    *title = 0;
  }
  if (length) {
    *length = -1;
  }

  // currently playing file
	if (!filename || !*filename) {
    d = app.cur_disk;
  } else {
    d = SC68app_load_diskfile(filename, 0);
  }
  if (d) {
    _snprintf(title, MAX_PATH, "%s - %s [%02u]", d->mus[d->default_six].author,
              d->name, d->nb_six);

    if (length) {
      *length = d->time * 1000;
    }
    if (d != app.cur_disk) {
      SC68_free(d);
    }
  }
}

void eq_set(int on, char data[10], int preamp)
{
	// most plug-ins can't even do an EQ anyhow.. I'm working on writing
	// a generic PCM EQ, but it looks like it'll be a little too CPU
	// consuming to be useful :)
}


// render 576 samples into buf.
// note that if you adjust the size of sample_buffer, for say, 1024
// sample blocks, it will still work, but some of the visualization
// might not look as good as it could. Stick with 576 sample blocks
// if you can, and have an additional auxiliary (overflow) buffer if
// necessary..
int get_samples(char *buf, int max)
{
	int sended = 0;
  const unsigned int sign = SC68MIXER_CHANGE_SIGN;
  unsigned int cycle_this_pass = app.mix.cycleperpass;

  if(!app.ready || !app.cur_mus) {
    return 0;
  }

  while (max>0) {
    int tosend;

    if( ! app.mix.buflen ) {
#ifdef _WRAPPER_
      EMU68_cycle(cycle_this_pass);
#else
      EMU68_level_and_interrupt(cycle_this_pass);
#endif


      /* YM ??? */
      if (app.cur_mus->flags.ym)
        app.mix.buflen = YM_mix(cycle_this_pass);
      else
        app.mix.buflen = app.mix.stdbuflen;

      app.mix.buf = YM_get_buffer();

      /* MW ??? */
      if (app.cur_mus->flags.ste) {
        MW_mix(app.mix.buf, app.reg68.mem, app.mix.buflen);
        if (sign) {
          SC68mixer_stereo_16_LR(app.mix.buf,app.mix.buf, app.mix.buflen, sign);
        }
      }
      /* Paula ??? */
      else if (app.cur_mus->flags.amiga) {
        PL_mix(app.mix.buf, app.reg68.mem, app.reg68.mem + app.reg68.memsz,
               app.mix.buflen);

        SC68mixer_stereo_16_LR(app.mix.buf,app.mix.buf, app.mix.buflen,
                               SC68MIXER_CHANGE_SIGN);

        SC68mixer_blend_LR(app.mix.buf, app.mix.buf, app.mix.buflen,
                           app.mix.amiga_blend , sign);
      }
      /* Do stereo my self */
      else {
        SC68mixer_dup_L_to_R(app.mix.buf, app.mix.buf, app.mix.buflen, sign);
      }

      if (app.delay.onoff) {
        /* Apply delay ... */
        int delay;
        const int shift = 0;

        if (app.delay.time < 0)
          delay = (app.mix.buflen << shift) / -app.delay.time;
        else
          delay = (app.mix.buflen * app.delay.time) >> shift;
        if (!app.delay.lr)
          /* Delay ring */
          SC68delay(app.mix.buf, app.mix.buflen, delay, app.delay.strength, sign);
        else
          /* L/R spacializer delay */
          SC68delay_LR(app.mix.buf, app.mix.buflen, delay);
      }


      /* Advance time */
      SC68app_advance_time(&app);

      /* Reach end of track */
      if (app.time.elapsed > app.time.track) {
        if(app.cur_track+1 >= app.cur_disk->nb_six) {
          memset(buf,0,max*4);
          return -1;
        } else {
          SC68track(&app, app.cur_track+1);
        }
      }
    }

    tosend = (app.mix.buflen > max) ? max : app.mix.buflen;
    memcpy(buf, app.mix.buf, tosend * 4);
    buf += tosend * 4;
    max -= tosend;
    app.mix.buf += tosend;
    app.mix.buflen -= tosend;
    sended += tosend;
  }
  return sended;
}

static char * get_sample_buffer ( int newSize )
{
  const int guardBand = 1024;
  newSize <<= 2;  // In bytes...
  if (newSize > sample_buffer_sz) {
    sample_buffer = realloc(sample_buffer, newSize + guardBand);
    if (!sample_buffer) {
      newSize = 0;
    }
    sample_buffer_sz = newSize;
  }
  return sample_buffer;
}

DWORD WINAPI __stdcall DecodeThread(void *b)
{
	int done=0;
  int decode_pos_ms = 0;

  // While not killed by stop()
	while (! *((int *)b))  {
    if(!app.ready) {
      Sleep(20);
      continue;
    }

    // Safe net
    if (!mod.outMod) {
      done = 1;
    }

		if (done) {
			if (!mod.outMod || !mod.outMod->IsPlaying()) {
				PostMessage(mod.hMainWindow,WM_WA_MPEG_EOF,0,0);
				return 0;
			}
			Sleep(10);
    } else {
      int dspActive   = !!mod.dsp_isactive();
      int sampleNeed  = 576 << dspActive;
      int missing;
      char *buffer;
      int read, dspRead;

      missing = (sampleNeed << 2) - mod.outMod->CanWrite();
      if (missing > 0) {
        missing >>= 2;  // In sample
        missing *= 10;
        missing /= 441; // Missing in Ms
        Sleep((missing >> 1) + 1); // Wait half of this time
        continue;
      }

      buffer = get_sample_buffer(sampleNeed);
      if (!buffer) {
        // Error allocating buffer !!
        done = 1;
        continue;
      }

      /* Getting samples from SC68 */
			read = get_samples(sample_buffer, 576);
			if (read < 0) {
        // SC68 error
        done = 1;
        continue;
			}

      if (dspActive) {
        dspRead = mod.dsp_dosamples((short *)sample_buffer, read, 16, 2, 44100);
      } else {
        dspRead = read;
      }
      if ((unsigned int)dspRead*4 > (unsigned int)sample_buffer_sz) {
        MessageBox(mod.hMainWindow,
          "Winamp DSP causes a buffer overflow\n"
          "Please report this bug <bug@sashipa.com>\n",
          "winamp68 error report", MB_OK | MB_ICONERROR);
        done = 1;
        continue;
      }

      // Get timestamp before writting this block.
      decode_pos_ms = mod.outMod->GetWrittenTime();

      // Send final DSP data to output module
			mod.outMod->Write(sample_buffer, dspRead << 2);

      // Simple vis.
			mod.SAAddPCMData ((char *)sample_buffer, 2, 16, decode_pos_ms);
			mod.VSAAddPCMData((char *)sample_buffer, 2, 16, decode_pos_ms);

      /* Local Timestamp for visual FX (FIX8) */
      /* $$$ not really sure for read/dspRead in following formula */
			//decode_pos_ms += (read*10) / 441;


		}
	}
	return 0;
}

In_Module mod =
{
	IN_VER,
#ifdef _DEBUG
	"DEBUG VERSION of sashipa Atari-ST & Amiga music player (v" WINAMP68_VERSION_STR ")"
#else
	"sashipa Atari-ST & Amiga music player (v" WINAMP68_VERSION_STR ")"
#endif
#ifdef __alpha
	"(AXP)"
#else
	"(x86)"
#endif
	,
	0,	// hMainWindow
	0,  // hDllInstance
	"sc68\0SC68 Audio File (*.sc68)\0"
	,
	1,	// is_seekable
	1,  // uses output
	config, //OK
	about,  //OK
	init,   //OK
	quit,   //OK
	getfileinfo,  //OK
	infoDlg,      //OK
	isourfile,    //OK
	play,         //OK
	pause,        //OK
	unpause,      //OK
	ispaused,     //OK
	stop,         //OK

	getlength,      //OK
	getoutputtime,  //OK
	setoutputtime,  //OK

	setvolume,  //OK
	setpan,     //OK

	0,0,0,0,0,0,0,0,0, // vis stuff


	0,0, // dsp

	eq_set, //?

	NULL,		// setinfo

	0 // out_mod

};

__declspec( dllexport ) In_Module * winampGetInModule2()
{
	return &mod;
}
